;==============================================================================
; (c) Copyright Elect Software International Inc., 1992, Toronto. Anyone can
; use this code for anything as long as it is not resold as a software
; development resource, as long as the copyright notice isn't removed, as
; long as changes are clearly marked as to authorship, and as long as users
; indemnify Elect from any liability.
; Comments welcome. Henrik Bechmann, CIS:72701,3717; Tel:416-534-8176.
;==============================================================================

;==============================================================================
;                                MENUBOX INTERFACE
;==============================================================================
; Description:
;     MenuBox provides a vertical bounce bar menu with optional vertical
;     scrolling.
;Creation and destruction...
;    MenuBox.Constructor(NumberOfRows,MenuBoxLength,MenuBoxTitle)
;        ; ... returns registration number, must be paired with Destructor()
;    ; must add exactly NumberOfRows items and dividers...
;    MenuBox.Destructor() ; must be paired with Constructor()
;    MenuBox.AddItem(SelectionValue) ; returns SelectionNumber
;    MenuBox.AddDivider() ; returns RowNumber
;Properties...
;   MenuBox.ResponseProc         ; default ""
;   MenuBox.Looping              ; default false
;   MenuBox.RefreshAfterResponse ; default true
;   MenuBox.SelectionValue       ; initialized to ""
;   MenuBox.SelectionNum         ; default 1
;   MenuBox.LaunchFromChar       ; default false
;   MenuBox.Footnote             ; default Box.MenuBoxFootnote

;   MenuBox.Keycode              ; initialized to blanknum()
;   MenuBox.KeyAccept            ; default Box.MenuBoxKeyAccept
;   MenuBox.KeyTop               ; default Box.MenuBoxKeyTop
;   MenuBox.KeyBottom            ; default Box.MenuBoxKeyBottom

;   MenuBox.TitleColors          ; default Box.MenuBoxTitleColors
;   MenuBox.BodyColors           ; default Box.MenuBoxBodyColors
;   MenuBox.BorderColors         ; default Box.MenuBoxBorderColors
;   MenuBox.ShadowColors         ; default Box.MenuBoxShadowColors
;   MenuBox.HighLightColors      ; default Box.MenuBoxHighlightColors
;Usage...
;    MenuBox.Display(RowCentreOffset,ColCentreOffset)
;        ; ... returns true if selection, false otherwise
;Query...
;    MenuBox.GetSelectionNumber() ; derives numeric value of selection from
;        ; current MenuBox.SelectionValue

;==============================================================================
;                           MENUBOX IMPLEMENTATION
;==============================================================================

Proc MenuBox.Constructor(NumberOfRows,MenuBoxLength,MenuBoxTitle)
   Private
      FrameShift,
      i
   If Box.MenuBoxIsAlive Then
      MenuBox!PushOnStack()
      MenuBox_PushedMenuBox = True
   Else
      MenuBox_PushedMenuBox = False
   Endif

   Global.Message = ""

   Array MenuBox_Items[NumberOfRows]
   Array MenuBox_ItemAttributes[NumberOfRows]
   For i From 1 to NumberOfRows
      MenuBox_ItemAttributes[i] = Blanknum()
   Endfor
   MenuBox_DividerAttribute = 0
   MenuBox_nDividers = 0
   MenuBox.SelectionNum = 1
   MenuBox.SelectionValue = ""
   MenuBox.Keycode = BlankNum()
   MenuBox_KeyBuffer = BlankNum()
   ; Make arguments available to MenuBox.Display(..)...
   MenuBox_nItems       = NumberOfRows
   MenuBox_Length       = MenuBoxLength
   MenuBox_Title        = MenuBoxTitle
   ; The following three can be OverRiden with Supplementary messages...
   MenuBox.ResponseProc = ""
   MenuBox.RefreshAfterResponse = True
   MenuBox.Looping      = False

   ; Initialize configuration and Colors...
   MenuBox.Footnote          = Box.MenuBoxFootnote
   MenuBox.TitleColors       = Box.MenuBoxTitleColors
   MenuBox.BodyColors        = Box.MenuBoxBodyColors
   MenuBox.BorderColors      = Box.MenuBoxBorderColors
   MenuBox.ShadowColors      = Box.MenuBoxShadowColors
   MenuBox.HighLightColors   = Box.MenuBoxHighLightColors

   ; Calculate width (maximum string length)
   MenuBox_InnerWidth = 0 ; this is set in MenuBox.Display() after
                          ; MenuBox_Items[] has been set.
   ; Calculate length...
   If MenuBox_Title <> "" Then
      MenuBox_InnerLength = MenuBox_Length + 2
   Else
      MenuBox_InnerLength = MenuBox_Length
   Endif

   ; Initialize display frame...
   MenuBox_FrameFirstItem  = 1
   MenuBox_FrameLastItem   = Min(MenuBox_Length,MenuBox_nItems)
   MenuBox_FrameLength     = MenuBox_FrameLastItem - MenuBox_FrameFirstItem + 1

   ; Initialize window pointer...
   MenuBox_Windowptr = MenuBox.SelectionNum
   MenuBox_ItemBeingAdded = 0
   MenuBox.LaunchFromChar = False
   MenuBox.KeyAccept    = Box.MenuBoxKeyAccept
   MenuBox.KeyTop       = Box.MenuBoxKeyTop
   MenuBox.KeyBottom    = Box.MenuBoxKeyBottom
   Return Box.RegisterType("MenuBox")
EndProc
Writelib    LibrarySpec    MenuBox.Constructor
Release     Procs          MenuBox.Constructor

Proc MenuBox.Destructor()
   Release Vars
      MenuBox.SelectionNum,
      MenuBox.SelectionValue,
      MenuBox.ResponseProc ,
      MenuBox.RefreshAfterResponse,
      MenuBox.Looping,
      MenuBox.Footnote,
      MenuBox.TitleColors,
      MenuBox.BodyColors,
      MenuBox.BorderColors,
      MenuBox.ShadowColors,
      MenuBox.HighLightColors,
      MenuBox.LaunchFromChar,
      MenuBox.Keycode,
      MenuBox_KeyBuffer,
      MenuBox.KeyAccept,
      MenuBox.KeyTop,
      MenuBox.KeyBottom,
      MenuBox_Items,
      MenuBox_ItemAttributes,
      MenuBox_DividerAttribute,
      MenuBox_nDividers,
      MenuBox_nItems,
      MenuBox_Length,
      MenuBox_Title,
      MenuBox_InnerWidth,
      MenuBox_InnerLength ,
      MenuBox_FrameFirstItem,
      MenuBox_FrameLastItem,
      MenuBox_FrameLength,
      MenuBox_Windowptr,
      MenuBox_ItemBeingAdded
   Box.DeregisterType("MenuBox")
   If MenuBox_PushedMenuBox Then
      MenuBox!PopFromStack()
   Else
      Release Vars
         MenuBox_PushedMenuBox
   Endif
EndProc
Writelib    LibrarySpec    MenuBox.Destructor
Release     Procs          MenuBox.Destructor

Proc MenuBox.AddItem(SelectionValue)
   MenuBox_ItemBeingAdded = MenuBox_ItemBeingAdded + 1
   If MenuBox_ItemBeingAdded > MenuBox_nItems Then
      Return 0
   Endif
   MenuBox_InnerWidth = Max(MenuBox_InnerWidth,Len(SelectionValue))
   MenuBox_Items[MenuBox_ItemBeingAdded] = SelectionValue
   Return MenuBox_ItemBeingAdded - MenuBox_nDividers
EndProc
Writelib    LibrarySpec    MenuBox.AddItem
Release     Procs          MenuBox.AddItem

Proc MenuBox.AddDivider()
   MenuBox_ItemBeingAdded = MenuBox_ItemBeingAdded + 1
   If MenuBox_ItemBeingAdded > MenuBox_nItems Then
      Return 0
   Endif
   MenuBox_ItemAttributes[MenuBox_ItemBeingAdded] = MenuBox_DividerAttribute
   MenuBox_nDividers = MenuBox_nDividers + 1
   MenuBox_Items[MenuBox_ItemBeingAdded] = ""
   Return MenuBox_ItemBeingAdded
EndProc
Writelib    LibrarySpec    MenuBox.AddDivider
Release     Procs          MenuBox.AddDivider


; MenuBoxLength must be 22 or less.
; Assume constructor has just run and refreshed settings.
Proc MenuBox.Display(RowCentreOffset,ColCentreOffset)
   Private
      ; Box.Display arguments...
      BoxRow,
      BoxCol,
      ; Selection control and response...
      NextSelectionNum,
      SelectionNumShift,
      FrameShift,
      ; Process control...
      DisplayMenu,
      Redraw,
      ; Misc...
      IsSelection,
      ItemPtr,
      i,
      Variance,
      Found

   If MenuBox.SelectionValue = "" Then
      i = 0
      While i <> MenuBox.SelectionNum
         i = i + 1
         If MenuBox_ItemAttributes[i] = MenuBox_DividerAttribute Then
            MenuBox.SelectionNum = MenuBox.SelectionNum + 1
         Endif
      EndWhile
      MenuBox.SelectionValue = MenuBox_Items[MenuBox.SelectionNum]
   Else
      For i From 1 to MenuBox_nItems
         Found = (MenuBox_Items[i] = MenuBox.SelectionValue)
         If Found Then
            Quitloop
         Endif
      Endfor
      If Not Found Then
         MenuBox.SelectionNum = 1
      Else
         MenuBox.SelectionNum = i
      Endif
      MenuBox.SelectionValue = MenuBox_Items[MenuBox.SelectionNum]
   Endif
   If MenuBox.SelectionNum > MenuBox_FrameLastItem Then
      FrameShift = MenuBox.SelectionNum - MenuBox_FrameLastItem
      MenuBox_FrameFirstItem = MenuBox_FrameFirstItem + FrameShift
      MenuBox_FrameLastItem = MenuBox_FrameLastItem + FrameShift
   Endif
   ; Initialize window pointer...
   MenuBox_Windowptr = MenuBox.SelectionNum - MenuBox_FrameFirstItem + 1
   MenuBox_InnerWidth = Max(Len(MenuBox_Title),MenuBox_InnerWidth)
   MenuBox_InnerWidth = Max(Len(MenuBox.Footnote) + 2,MenuBox_InnerWidth)
   If MenuBox_InnerWidth + 5 > 80 Then
      MenuBox_InnerWidth = 75
   Endif
   MenuBox.SelectionValue = MenuBox_Items[MenuBox.SelectionNum]

   ; Centre the menu...
   BoxRow = int((25 - (MenuBox_InnerLength + 3))/2) + RowCentreOffset
   BoxCol = int((80 - (MenuBox_InnerWidth + 5))/2) + ColCentreOffset
   ; Display the box...
   Box.Display(BoxRow,BoxCol,MenuBox_InnerLength,MenuBox_InnerWidth,
               MenuBox_Title,MenuBox.Footnote,MenuBox.TitleColors,
               MenuBox.BodyColors,MenuBox.BorderColors,
               MenuBox.ShadowColors,MenuBox.HighLightColors)
   ; Display the menu...
   ReDraw = True
   DisplayMenu = True
   While DisplayMenu
      If Redraw Then
         ; redraw the menu item within the box...
         ItemPtr = MenuBox_FrameFirstItem
         For i From 1 to MenuBox_FrameLength
            If MenuBox_ItemAttributes[i] <> MenuBox_DividerAttribute Then
               Box.WriteRow(i,MenuBox_Items[ItemPtr])
            Else
               Box.WriteRow(i,Fill("",MenuBox_InnerWidth))
            Endif
            ItemPtr = ItemPtr + 1
         EndFor
         Box.HighlightRow(MenuBox_Windowptr)
         If MenuBox_FrameFirstItem > 1 Then
            Box.MoreAboveArrow(True)
         Else
            Box.MoreAboveArrow(False)
         Endif
         If MenuBox_FrameLastItem < MenuBox_nItems Then
            Box.MoreBelowArrow(True)
         Else
            Box.MoreBelowArrow(False)
         Endif
         ReDraw = False
      Endif
      ; take the user's response...
      MenuBox!ProcessKeyStrokes()
   EndWhile
   Return IsSelection
EndProc
Writelib    LibrarySpec    MenuBox.Display
Release     Procs          MenuBox.Display

Proc MenuBox!ProcessKeyStrokes()
   Private
      TestPtr
   MenuBox.Keycode = MenuBox_KeyBuffer
   MenuBox_KeyBuffer = BlankNum()
   If IsBlank(MenuBox.Keycode) Then
      If Global.Message <> "" Then
         Message Global.Message
      Endif
      MenuBox.Keycode = getchar()
      If Global.Message <> "" Then
         Global.Message = ""
         Box.ClearMessage()
      Endif
   Endif
   SelectionNumShift = 0
   Switch
      Case MenuBox.Keycode > 32: ; Search for first Char
         If MenuBox.LaunchFromChar Then
            NextSelectionNum = MenuBox!FindSelectChar()
            If NextSelectionNum <> 0 Then
               SelectionNumShift = NextSelectionNum - MenuBox.SelectionNum
               MenuBox_KeyBuffer = MenuBox.KeyAccept
            Endif
         Else
            NextSelectionNum = MenuBox!FindNextChar()
            SelectionNumShift = NextSelectionNum - MenuBox.SelectionNum
         Endif
      ; Navigation keys...
      Case MenuBox.Keycode = -72: ; Key$Up:
         If MenuBox.SelectionNum > 1 Then
            SelectionNumShift = -1
            While MenuBox_ItemAttributes[MenuBox.SelectionNum + SelectionNumShift]
               = MenuBox_DividerAttribute
               SelectionNumShift = SelectionNumShift - 1
            EndWhile
         Else
            MenuBox_KeyBuffer = MenuBox.KeyBottom
         Endif
      Case MenuBox.Keycode = -80: ; Key$Down:
         If MenuBox.SelectionNum < MenuBox_nItems Then
            SelectionNumShift = 1
            While MenuBox_ItemAttributes[MenuBox.SelectionNum + SelectionNumShift]
               = MenuBox_DividerAttribute
               SelectionNumShift = SelectionNumShift + 1
            EndWhile
         Else
            MenuBox_KeyBuffer = MenuBox.KeyTop
         Endif
      Case MenuBox.Keycode = MenuBox.KeyTop:
         SelectionNumShift = 1 - MenuBox.SelectionNum
      Case MenuBox.Keycode = MenuBox.KeyBottom:
         SelectionNumShift = MenuBox_nItems - MenuBox.SelectionNum
      Case MenuBox.Keycode = -73: ; Key$PgUp:
         SelectionNumShift = -MenuBox_Length
         If MenuBox.SelectionNum + SelectionNumShift < 1 Then
            SelectionNumShift = -(MenuBox.SelectionNum - 1)
         Endif
         While MenuBox_ItemAttributes[MenuBox.SelectionNum + SelectionNumShift]
            = MenuBox_DividerAttribute
            SelectionNumShift = SelectionNumShift - 1
         EndWhile
      Case MenuBox.Keycode = -81: ; Key$PgDn:
         SelectionNumShift = MenuBox_Length
         If MenuBox.SelectionNum + SelectionNumShift > MenuBox_nItems Then
            SelectionNumShift = MenuBox_nItems - MenuBox.SelectionNum
         Endif
         While MenuBox_ItemAttributes[MenuBox.SelectionNum + SelectionNumShift]
            = MenuBox_DividerAttribute
            SelectionNumShift = SelectionNumShift + 1
         EndWhile
      ; Select an item...
      Case MenuBox.Keycode = MenuBox.KeyAccept:
         IsSelection = True
         MenuBox.SelectionValue = MenuBox_Items[MenuBox.SelectionNum]
         MenuBox!ExecProc(MenuBox.ResponseProc)
         DisplayMenu = MenuBox.Looping
         If DisplayMenu Then
            Prompt "",""
            Canvas Off
            If MenuBox.RefreshAfterResponse Then
               Echo Normal
               Echo Off
            Endif
            Canvas On
            Prompt
            Box.Display(BoxRow,BoxCol,MenuBox_InnerLength,MenuBox_InnerWidth,
                        MenuBox_Title,MenuBox.Footnote,MenuBox.TitleColors,
                        MenuBox.BodyColors,MenuBox.BorderColors,
                        MenuBox.ShadowColors,MenuBox.HighLightColors)
            Redraw = true
         Endif
      ; Abandon the menu...
      Case MenuBox.Keycode = 27: ; Key$Esc:
         DisplayMenu = False
         IsSelection = False
   EndSwitch
   If SelectionNumShift <> 0 Then ; recalculate the display frame...
      Box.BodyColorRow(MenuBox_Windowptr)
      MenuBox.SelectionNum = MenuBox.SelectionNum + SelectionNumShift
      MenuBox.SelectionValue = MenuBox_Items[MenuBox.SelectionNum]
      TestPtr = MenuBox_Windowptr + SelectionNumShift
      If TestPtr < 1 Or TestPtr > MenuBox_Length Then
         ; Shift window...
         MenuBox_FrameFirstItem = MenuBox_FrameFirstItem + SelectionNumShift
         MenuBox_FrameLastItem = MenuBox_FrameLastItem + SelectionNumShift
         FrameShift = 0
         If MenuBox_FrameFirstItem < 1 Then
            FrameShift = -(MenuBox_FrameFirstItem - 1)
         Endif
         If MenuBox_FrameLastItem > MenuBox_nItems Then
            FrameShift = -(MenuBox_FrameLastItem - MenuBox_nItems)
         Endif
         If FrameShift <> 0 Then
            MenuBox_FrameFirstItem = MenuBox_FrameFirstItem + FrameShift
            MenuBox_FrameLastItem = MenuBox_FrameLastItem + FrameShift
            MenuBox_Windowptr = MenuBox.SelectionNum - MenuBox_FrameFirstItem + 1
         Endif
         Redraw = true
      Else ; Shift highlighted row only...
         MenuBox_Windowptr = TestPtr
         Box.HighlightRow(MenuBox_Windowptr)
      Endif ; window has to be shifted
   Endif ; there has been a change in selection
EndProc
Writelib    LibrarySpec    MenuBox!ProcessKeyStrokes
Release     Procs          MenuBox!ProcessKeyStrokes

; Search for the next occurence of the selection beginning with the character
; typed by the user...
Proc MenuBox!FindNextChar()
   Private
      SearchChar,
      i,
      Found
   Found = False
   SearchChar = Upper(Chr(MenuBox.Keycode))
   For i From MenuBox.SelectionNum + 1 To MenuBox_nItems
      Found = (SearchChar = Upper(Substr(MenuBox_Items[i],1,1)))
      If Found Then
         Return i
      Endif
   EndFor
   For i From 1 To MenuBox.SelectionNum
      Found = (SearchChar = Upper(Substr(MenuBox_Items[i],1,1)))
      If Found Then
         Return i
      Endif
   Endfor
   Return MenuBox.SelectionNum ; nothing found
EndProc
Writelib    LibrarySpec    MenuBox!FindNextChar
Release     Procs          MenuBox!FindNextChar

Proc MenuBox!FindSelectChar()
   Private
      SearchChar,
      SearchPos,
      i,
      Found
   Found = False
   SearchChar = Upper(Chr(MenuBox.Keycode))
   For i From MenuBox.SelectionNum + 1 To MenuBox_nItems
      SearchPos = Search(SearchChar,MenuBox_Items[i])
      Found = (SearchPos <> 0)
      If Found Then
         Found = (SearchChar = Substr(MenuBox_Items[i],SearchPos,1)) ; check for case
         If Found Then
            Return i
         Endif
      Endif
   EndFor
   For i From 1 To MenuBox.SelectionNum
      SearchPos = Search(SearchChar,MenuBox_Items[i])
      Found = (SearchPos <> 0)
      If Found Then
         Found = (SearchChar = Substr(MenuBox_Items[i],SearchPos,1)) ; check for case
         If Found Then
            Return i
         Endif
      Endif
   Endfor
   Return 0 ; nothing found
EndProc
Writelib    LibrarySpec    MenuBox!FindSelectChar
Release     Procs          MenuBox!FindSelectChar

Proc MenuBox.GetSelectionNumber()
   Private
      i,
      ActualSelectionNum
   ActualSelectionNum = MenuBox.SelectionNum
   If MenuBox_NDividers <> 0 Then
      For i From MenuBox.SelectionNum to 1 Step -1
         If MenuBox_ItemAttributes[i] = MenuBox_DividerAttribute Then
            ActualSelectionNum = ActualSelectionNum - 1
         Endif
      EndFor
   Endif
   Return ActualSelectionNum
EndProc
Writelib    LibrarySpec    MenuBox.GetSelectionNumber
Release     Procs          MenuBox.GetSelectionNumber

Proc MenuBox!PushOnStack()
   Message "Working..."
   SaveVars
      MenuBox.SelectionNum,
      MenuBox.SelectionValue,
      MenuBox.ResponseProc,
      MenuBox.RefreshAfterResponse,
      MenuBox.Looping,
      MenuBox.Footnote,
      MenuBox.TitleColors,
      MenuBox.BodyColors,
      MenuBox.BorderColors,
      MenuBox.ShadowColors,
      MenuBox.HighLightColors,
      MenuBox.LaunchFromChar,
      MenuBox.Keycode,
      MenuBox_KeyBuffer,
      MenuBox.KeyAccept,
      MenuBox.KeyTop,
      MenuBox.KeyBottom,
      MenuBox_PushedMenuBox,
      MenuBox_Items,
      MenuBox_ItemAttributes,
      MenuBox_DividerAttribute,
      MenuBox_nDividers,
      MenuBox_nItems,
      MenuBox_Length,
      MenuBox_Title,
      MenuBox_InnerWidth,
      MenuBox_InnerLength,
      MenuBox_FrameFirstItem ,
      MenuBox_FrameLastItem,
      MenuBox_FrameLength,
      MenuBox_Windowptr,
      MenuBox_ItemBeingAdded
   {Tools}{Rename}{Script}
      Select "savevars"
      Select Box.VirtualDrive + "BX$MB" +
         Substr(Format("W4,EZ",Box.MenuBoxNumber),2,3)
   If MenuChoice() = "Cancel" Then
      {Replace}
   Endif
   Box.MenuBoxIsAlive = False
EndProc
Writelib    LibrarySpec    MenuBox!PushOnStack
Release     Procs          MenuBox!PushOnStack

Proc MenuBox!PopFromStack()
   Message "Working..."
   Play "BX$MB" +
      Substr(Format("W4,EZ",Box.MenuBoxNumber),2,3)
   {Tools}{Delete}{Script}
      Select Box.VirtualDrive + "BX$MB" +
         Substr(Format("W4,EZ",Box.MenuBoxNumber),2,3)
   {OK}
   Box.MenuBoxIsAlive = True
EndProc
Writelib    LibrarySpec    MenuBox!PopFromStack
Release     Procs          MenuBox!PopFromStack

Proc MenuBox!ExecProc(ProcName)
   If IsBlank(ProcName) Then
      Return True
   Else
      ExecProc ProcName
      Return Retval
   Endif
EndProc
Writelib    LibrarySpec    MenuBox!ExecProc
Release     Procs          MenuBox!ExecProc